#include "FxCyanRGB8.hpp"

// C++ system level
#include <string>
#include <cstring>  // memset
#include <algorithm>
#include <functional>

// ESP32 specific
// #include "freertos/FreeRTOS.h"
// #include "freertos/queue.h"

// project specific
#include <Types.hpp>
#include "FxVSync.hpp"
#include "CyanBus.hpp"

// qthing stuff
// #include ""



namespace SiliconTorch {

  namespace FxCyanRGB8 {

    const std::string HEADER("fxCyanRGB8");


    FxCyanRGB8::FxCyanRGB8(u16 leds, u16 startIdx) : leds(leds), startIdx(startIdx) {

      buffer0 = new qthing::RGB[leds];
      buffer1 = new qthing::RGB[leds];

      // std::memset(buffer0, 0x00, leds * sizeof(Pixel));
      // std::memset(buffer1, 0x00, leds * sizeof(Pixel));



    }

    void FxCyanRGB8::handlePacket(u8* buffer, std::size_t length) {

      if (length < 3) {
        ESP_LOGE(HEADER.c_str(), "Wrong |buffer|[ %d ]: Must be greater than 2!", length);
        return;
      }

      if ((length - 2) % 3 > 0) {
        ESP_LOGE(HEADER.c_str(), "Wrong |buffer|[ %d ]: (Length - 2) must be divisible by 3!", length);
        return;
      }


      Pixel* packetPx = reinterpret_cast<Pixel*>(buffer+2);
      u16 packetPxs = (length - 2) / 3;
      u16 packetStartIdx = (buffer[0] << 8) | buffer[1];


      if (packetStartIdx == startIdx) {

        for (u16 px = 0; px < std::min(leds, packetPxs); px++) {
          buffer1[px].r = packetPx[px].r / 255.0f;
          buffer1[px].g = packetPx[px].g / 255.0f;
          buffer1[px].b = packetPx[px].b / 255.0f;
        }

      } else if(packetStartIdx > startIdx) {

        u16 delta = packetStartIdx - startIdx;

        for (u16 px = delta; px < std::min(leds, (u16)(packetPxs + delta)); px++) {
          buffer1[px].r = packetPx[px - delta].r / 255.0f;
          buffer1[px].g = packetPx[px - delta].g / 255.0f;
          buffer1[px].b = packetPx[px - delta].b / 255.0f;
        }

      } else if (packetStartIdx < startIdx) {

        u16 delta = startIdx - packetStartIdx;

        // Packet does not contain any pixels for us
        if (packetPxs < delta) return;

        for (u16 px = 0; px < std::min(leds, (u16)(packetPxs - delta)); px++) {
          buffer1[px].r = packetPx[px + delta].r / 255.0f;
          buffer1[px].g = packetPx[px + delta].g / 255.0f;
          buffer1[px].b = packetPx[px + delta].b / 255.0f;
        }
      }
    }

    void FxCyanRGB8::registerAtCyanBus(CyanBus::CyanBus& cyanBus) {

      u8 headerLength = HEADER.length();

      cyanBus.packetCallback.add(HEADER, [&, headerLength](const SiliconTorch::CyanBus::PacketData& packet) {
        this->handlePacket((u8*)(packet.payload + headerLength), packet.length - headerLength);
      });

      FxVSync::registerAtCyanBus(cyanBus, [&]() {
        requestBufferSwap();
      });

    }

    void FxCyanRGB8::registerUDPHandler() {
      if (!udpHandlerRegistered) {
        udpHandlerRegistered = true;

        u8 headerLength = HEADER.length();

        qthing::addUDPPacketCallback(HEADER, [&, headerLength](qthing::udpPacket packet) {
          this->handlePacket((u8*)(packet.payload + headerLength), packet.length - headerLength);
          requestBufferSwap();
        });
      }
    }

    // qthing::Animator
    qthing::RGB FxCyanRGB8::render(u16 addr) {
      if (addr < leds) {
        return buffer0[addr];
      } else {
        qthing::RGB black;
        return black;
      }
    }

    // qthing::Animator
    void FxCyanRGB8::step() {
      if (swapRequested) swapBuffers();
    }

    void FxCyanRGB8::requestBufferSwap() {
      swapRequested = true;
    }

    void FxCyanRGB8::swapBuffers() {
      swapRequested = false;
      qthing::RGB* tmp = buffer0;
      buffer0 = buffer1;
      buffer1 = tmp;

      for (u16 px = 0; px < leds; px++){
        buffer1[px].r = 0;
        buffer1[px].g = 0;
        buffer1[px].b = 0;
      }
    }


    str colorOrder2String(ColorOrder co) {
      switch (co) {
        case ColorOrder::RGB: return "RGB";
        case ColorOrder::RBG: return "RBG";
        case ColorOrder::GRB: return "GRB";
        case ColorOrder::GBR: return "GBR";
        case ColorOrder::BRG: return "BRG";
        case ColorOrder::BGR: return "BGR";
      }

      return "RGB";
    }

    ColorOrder string2ColorOrder(str co) {
      if      (co.compare("RGB") == 0) return RGB;
      else if (co.compare("RBG") == 0) return RBG;
      else if (co.compare("GRB") == 0) return GRB;
      else if (co.compare("GBR") == 0) return GBR;
      else if (co.compare("BRG") == 0) return BRG;
      else if (co.compare("BGR") == 0) return BGR;

      return ColorOrder::RGB;
    }

  }
}
