#include "Service.hpp"

// C++ system level
// #include <cstring>     // memset, strncmp
// #include <cstdlib>     // TODO: is this for memcpy?
// #include <functional>

// ESP32 specific
#include "esp_log.h"

// project specific
#include <Types.hpp>
#include "SiliconTorch/NVSExplorer.hpp"

// qthing stuff
// #include <qthing>


static const char* TAG = "ServiceManager";


namespace SiliconTorch {

  namespace ServiceManager {

    const str NVSStateKey = "state";


    Service::Service() {
      // u8 state = 0;  // disabled -> 0   enabled -> 1
      // state = nvs.getUnsignedInt(nameSpace(), NVSStateKey, state);

      // if (state >= (u8)Invalid) state = (u8)Disabled;
      // nvsState = (NVSState)state;
    }

    void Service::preInit(ServiceLookup lookup) {
      serviceLookup = lookup;
    }

    void Service::postInit() {

      attrsLocked = true;


      for (const auto& name : *wantedNames) {
        Service* service = otherByName(name);

        if (service != NULL) wantedServices.insert(service);
      }

      for (const auto& name : *requiredNames) {
        Service* service = otherByName(name);

        if (service != NULL) {
          requiredServices.insert(service);
        } else {
          hasErrors = true;  // service is errorneous because of unsatisfied dependency
          ESP_LOGW(TAG, "Service[ %s ] is errorneous because of unsatisfied dependency[ %s ]", getName().c_str(), name.c_str());
        }
      }

      delete wantedNames;
      delete requiredNames;

      wantedNames = NULL;
      requiredNames = NULL;


      u8 state = 0;  // disabled -> 0   enabled -> 1
      state = SiliconTorch::NVSExplorer::NVSExplorer::instance().getUnsignedInt(getNameSpace(), NVSStateKey, state);

      if (state >= (u8)Invalid) state = (u8)Disabled;
      nvsState = (NVSState)state;

    }

    bool Service::isEnabled() const {
      return nvsState == Enabled && !hasErrors;
    }

    void Service::setEnabled(bool enabled) {

      NVSState state = enabled ? Enabled : Disabled;
      u8 stateValue = (u8)state;

      bool result = SiliconTorch::NVSExplorer::NVSExplorer::instance().setU8(getNameSpace(), NVSStateKey, stateValue);

      if (!result) {
        SiliconTorch::NVSExplorer::NVSExplorer::instance().removeKey(getNameSpace(), NVSStateKey);
        result = SiliconTorch::NVSExplorer::NVSExplorer::instance().setU8(getNameSpace(), NVSStateKey, stateValue);
      }

      if (result) {
        nvsState = state;
        ESP_LOGI(TAG, "Setting service[ %s ] to %s", getName().c_str(), enabled ? "enabled[✅]" : "disabled[❌]");
      } else {
        ESP_LOGW(TAG, "Error saving service[ %s] state!", getName().c_str());
      }
    }


    void Service::setName(const str& name) {
      if (!attrsLocked) {
        this->name = name;
      } else {
        // TODO: report error somehow?
      }
    }

    void Service::setNameSpace(const str& nameSpace) {
      if (!attrsLocked) {
        this->nameSpace = nameSpace;
      } else {
        // TODO: report error somehow?
      }
    }


    void Service::addWantedService(const str& name) {
      if (!attrsLocked) {
        wantedNames->insert(name);
      } else {
        // TODO: report error somehow?
      }
    }

    void Service::addRequiredService(const str& name) {
      if (!attrsLocked) {
        requiredNames->insert(name);
      } else {
        // TODO: report error somehow?
      }
    }


    ServiceSet Service::getWantedServices() {
      ServiceSet out;

      // TODO: better method for set copy?!

      for (const auto& service : wantedServices)
        out.insert(service);

      return out;
    }

    ServiceSet Service::getRequiredServices() {
      ServiceSet out;

      // TODO: better method for set copy?!

      for (const auto& service : requiredServices)
        out.insert(service);

      return out;
    }


    Service* Service::otherByName(const str& name) {
      return serviceLookup(name);
    }


    const str& Service::getName() const {
      return name;
    }

    const str& Service::getNameSpace() const {
      return nameSpace;
    }

  }
}
